/*****************************
* Source : pgsql.prg
* System : <unkown>
* Author : Phil Ide
* Created: 02/02/2004
*
* Purpose:
* ----------------------------
* History:
* ----------------------------
*    02/02/2004 14:19 PPI - Created
*****************************/

/*
  For an explanation of how to this works, see:

  C:\Program Files\PostGreSQL\usr\doc\postgresql-7.3.1\html\programmer-client.html

  (assuming you installed to the default directories)
*/

#include "Error.ch"
#include "piCommon.ch"
#include "dll.ch"

#xcommand DEFAULT <var1> TO <val1> [,<varN> TO <valN>];
       => IF <var1> == NIL ; <var1> := <val1> ; ENDIF ;
       [; IF <varN> == NIL ; <varN> := <valN> ; ENDIF ]


#define TRUE .T.
#define FALSE .F.

#pragma Map(PGDeref,"_XBPPGDEREF")


CLASS PGSql
   EXPORTED:
      VAR connID READONLY
      VAR result

      METHOD init
      METHOD connect
      METHOD setdbLogin
      METHOD disconnect
      METHOD reset

      // non-blocking connection
      METHOD connectStart
      METHOD connectPoll
      METHOD resetStart
      METHOD resetPoll

      METHOD setnonblocking
      METHOD isnonblocking
      METHOD sendQuery
      METHOD getResult
      METHOD consumeInput
      METHOD isBusy
      METHOD flush
      METHOD socket
      METHOD requestCancel

      // trace/debug
      METHOD trace
      METHOD untrace


      METHOD exec
      METHOD execParams
      METHOD escapeString
      METHOD escapeBytea
      METHOD unescapeBytea

      METHOD makeEmptyPGResult

      // info
      METHOD db
      METHOD user
      METHOD pass
      METHOD host
      METHOD port
      METHOD options
      METHOD status
      METHOD errorMessage
      METHOD backendPID
ENDCLASS

METHOD PGSql:init()
   ::connID := 0
   return self

METHOD PGSql:setdbLogin(cHost,cPort,cOpts,cTty, cDbName,cUser,cPwd)
   local lOk

   ::connID := PQsetdbLogin(cHost,cPort,cOpts,cTty, cDbName,cUser,cPwd)
   if !( lOk := (PQstatus(::connID) == 0) )
      ::connID := 0
   endif
   return lOk

METHOD PGSql:connect( cConnString )
   local lOk

   ::connID := PQconnectdb(cConnString)
   if !( lOk := (PQstatus(::connID) == 0) )
      ::connID := 0
   endif
   return lOk

METHOD PGSql:connectStart( cConnString )
   local lOk

   ::connID := PQconnectdb(cConnString)
   if !( lOk := (PQstatus(::connID) == 0) )
      ::connID := 0
   endif
   return lOk

METHOD PGSql:connectPoll()
   return PQconnectPoll(::connID)

METHOD PGSql:resetStart()
   return PQresetStart(::connID)

METHOD PGSql:resetPoll()
   return PQresetPoll(::connID)

METHOD PGSql:disconnect()
   PQfinish(::connID)
   return self

METHOD PGSql:reset()
   local lOk

   PQreset(::connID)
   if !( lOk := (PQstatus(::connID) == 0) )
      ::connID := 0
   endif
   return lOk

METHOD PGSql:escapeString(c)
   local cBuff := Space( (Len(c)*2)+1 )
   PQescapeString( @cBuff, c, Len(c) )
   return cBuff

METHOD PGSql:escapeBytea(c)
   local cBuff
   local i := 0

   cBuff := PQescapeBytea( c, Len(c), @i )
   return PGDeref(cBuff)

METHOD PGSql:unescapeBytea(c)
   local cBuff
   local i := 0

   cBuff := PQunescapeBytea( c, Len(c) )
   return PGDeref(cBuff)



METHOD PGSql:exec(cSQL)
   local xT
   local oQ
   local lOk := TRUE

   ::result := NIL
   xT := PQexec(::connID,cSQL)
   if xT == 0
      lOk := FALSE
   else
      ::result := PGResult():new( self, xT )
   endif
   return lOk

METHOD PGSql:execParams(cSql,nParams,ct,cv,cl,cf,rf)
   local xT
   local oQ
   local lOk := TRUE

   ::result := NIL
   xT := PQexecParams(::connID,cSql,nParams,ct,cv,cl,cf,rf)
   if xT == 0
      lOk := FALSE
   else
      ::result := PGResult():new( self, xT )
   endif
   return lOk


METHOD PGSql:db()
   local c := PQdb(::connID)
   c := PGDeref(c)
   return c

METHOD PGSql:user()
   local c := PQuser(::connID)
   c := PGDeref(c)
   return c

METHOD PGSql:pass()
   local c := PQpass(::connID)
   c := PGDeref(c)
   return c

METHOD PGSql:host()
   local c := PQhost(::connID)
   c := PGDeref(c)
   return c

METHOD PGSql:port()
   local c := PQport(::connID)
   c := PGDeref(c)
   return c

METHOD PGSql:options()
   local c := PQoptions(::connID)
   c := PGDeref(c)
   return c

METHOD PGSql:status()
   return PQstatus(::connID)

METHOD PGSql:errorMessage()
   local c := PQerrorMessage(::connID)
   c := PGDeref(c)
   return c

METHOD PGSql:backendPID()
   return PQbackendPID(::connID)

METHOD PGSql:makeEmptyPGResult(i)
   local x := PQmakeEmptyPGResult(::connID,i)
   local o

   o := PGResult():new(::connID,x)
   return o


METHOD PGSql:setnonblocking(lNonBlockingMode)
   return PQsetnonblocking(::connID,iif(lNonBlockingMode,1,0)) == 0

METHOD PGSql:isnonblocking()
   return PQisnonblocking(::connID) == 1

METHOD PGSql:sendQuery( cQuery )
   return PQsendQuery(::connID, cQuery) == 1

METHOD PGSql:getResult
   local o := PQgetResult(::connID)
   o := PGResult():new( self, o )
   return o

METHOD PGSql:consumeInput()
   return PQconsumeInput(::connID)

METHOD PGSql:isBusy
   return PQisBusy(::connID) == 1

METHOD PGSql:flush()
   return PQflush(::connID) == 0

METHOD PGSql:socket()
   return PQsocket(::connID) // >0 = connector, <0 is error (-1 no backend!)

METHOD PGSql:requestCancel()
   return PQrequestCancel(::connID) == 1

/*
:requestCancel()
The return value is 1 if the cancel request was successfully dispatched, 0 if not.
(If not, PQerrorMessage tells why not.) Successful dispatch is no guarantee that the
request will have any effect, however. Regardless of the return value of
PQrequestCancel, the application must continue with the normal result-reading
sequence using PQgetResult. If the cancellation is effective, the current command
will terminate early and return an error result. If the cancellation fails (say,
because the backend was already done processing the command), then there will be no
visible result at all.
*/

METHOD PGSql:trace( nFilehandle )
   PQtrace(::connID, nFilehandle )
   return self

METHOD PGSql:untrace()
   PQuntrace(::connID)
   return self



//<!===================================================!>//
CLASS PGResult
   EXPORTED:
      VAR resID READONLY
      VAR curTuple READONLY
      VAR connID
      VAR rows
      VAR cols
      VAR message


      METHOD init
      METHOD destroy
      METHOD clear
      METHOD fname
      METHOD fnumber
      METHOD ftype
      METHOD fmod
      METHOD fsize
      METHOD getValue
      METHOD getLength
      METHOD getIsNULL
      METHOD binaryTuples

      METHOD cmdStatus
      METHOD cmdTuples
      METHOD oidValue
      METHOD oidStatus

      METHOD resultStatus
      METHOD resStatus
      METHOD resultErrorMessage
      METHOD makeEmptyPGResult

      METHOD escapeString
      METHOD escapeBytea
      METHOD unescapeBytea

      ACCESS ASSIGN METHOD NoIvarCallBack

      METHOD skip
      METHOD goTop
      METHOD goTo
      METHOD goBottom
      METHOD eof
      METHOD bof
      METHOD recno
      METHOD position
      METHOD goPosition

      METHOD getFieldNames

   PROTECTED:
      METHOD NoIVar
      VAR eofMark READONLY
      VAR bofMark READONLY

ENDCLASS

METHOD PGResult:init( oConn, resID )
   ::connID := oConn
   ::resID  := resID
   ::rows := PQntuples(::resID)
   ::cols := PQnfields(::resID)
   ::curTuple := 0
   ::eofMark := (::rows == 0)
   ::bofMark := (::rows == 0)
   return self

METHOD PGResult:resultStatus()
   return PQresultStatus(::resID)

METHOD PGResult:resStatus(i)
   return PGDeref(PQresStatus(i))

METHOD PGResult:makeEmptyPGResult(i)
   return ::connID:makeEmptyPGResult(i)


METHOD PGResult:resultErrorMessage()
   return PGDeref(PQresultErrorMessage(::resID))


METHOD PGResult:eof()
   return ::eofMark

METHOD PGResult:bof()
   return ::bofMark

METHOD PGResult:recno()
   return ::curTuple

METHOD PGResult:position()
   local nPC := ::rows/100
   return (::curTuple+1)/nPC

METHOD PGResult:goPosition(n)
   local nPC := ::rows/100

   ::curTuple := int(nPC*n)
   return ::curTuple


METHOD PGResult:skip(n)
   default n to 1
   ::curTuple += n
   do case
      case ::curTuple < 0
         ::curTuple := 0
         ::bofMark := TRUE
         ::eofMark := FALSE

      case ::curTuple > ::rows-1
         ::curTuple := ::rows-1
         ::eofMark := TRUE
         ::bofMark := FALSE

      otherwise
         ::eofMark := FALSE
         ::bofMark := FALSE
   endcase
   return self

METHOD PGResult:goTop()
   ::curTuple := 0
   ::eofMark := FALSE
   ::bofMark := FALSE
   return self


METHOD PGResult:goTo(n)
   if ValType(n) == 'N'
      if n >= 0 .and. n < ::cols
         ::curTuple := n
         ::eofMark := FALSE
         ::bofMark := FALSE

      elseif n < 0
         ::bofMark := TRUE
         ::eofMark := FALSE

      elseif n > ::cols
         ::eofMark := TRUE
         ::bofMark := FALSE

      endif
   endif
   return ::curTuple

METHOD PGResult:goBottom()
   ::curTuple := ::rows-1
   ::eofMark := FALSE
   ::bofMark := FALSE
   return self

METHOD PGResult:getFieldNames()
   local i
   local aRet := Array(::cols)

   for i := 0 to ::cols-1
      aRet[i+1] := ::fname(i)
   next
   return aRet

METHOD PGResult:NoIvarCallBack(x)
   local i
   local xRet

   i := PQfnumber(::resID, ::message)
   if i >= 0
      xRet := ::getValue(::curTuple, i)
   else
      ::noIvar(x)
   endif
   return xRet

METHOD PGResult:NoIvar( x )
   local o := Error():new()

   o:args            := x
   o:canDefault      := FALSE
   o:canRetry        := FALSE
   o:canSubstitute   := FALSE
   o:description     := "invalid field name"
   o:operation       := ::className()+":"+::message
   o:severity        := XPP_ES_ERROR
   o:subSystem       := ::className()
   o:thread          := ThreadID()
   Eval( ErrorBlock(), o )
   return self


METHOD PGResult:destroy()
   if !(::resID == NIL)
      PQclear(::resID)
      ::resID := NIL
   endif
   return self

METHOD PGResult:clear()
   ::destroy()
   return self

METHOD PGResult:fname(i)
   local c := PQfname(::resID, i)
   return PGDeref(c)

METHOD PGResult:fnumber(cField)
   return PQfnumber(::resID, cField)

METHOD PGResult:ftype(nField)
   return PQftype(::resID, nField)

METHOD PGResult:fmod(nField)
   return PQfmod(::resID, nField)

METHOD PGResult:fsize(nField)
   return PQfsize(::resID, nField)

METHOD PGResult:getValue( nRow, nCol )
   local c := PQgetvalue(::resID,nRow,nCol)
   return PGDeref(c)

METHOD PGResult:binaryTuples()
   return PQbinaryTuples(::resID) == 1

METHOD PGResult:getLength( nRow, nCol )
   return PQgetlength(::resID,nRow,nCol)

METHOD PGResult:getIsNULL( nRow, nCol )
   return PQgetisnull(::resID,nRow,nCol) == 1

METHOD PGResult:cmdStatus()
   local c := PQcmdStatus(::resID)
   return PGDeref(c)

METHOD PGResult:cmdTuples()
   local c := PQcmdTuples(::resID)
   return PGDeref(c)

METHOD PGResult:oidValue()
   return PQoidValue(::resID)

METHOD PGResult:oidStatus()
   local c := PQoidStatus(::resID)
   return PGDeref(c)

METHOD PGResult:escapeString(c)
   return ::connID:escapeString(c)

METHOD PGResult:escapeBytea(c)
   return ::connID:escapeBytea(c)

METHOD PGResult:unescapeBytea(c)
   return ::connID:unescapeBytea(c)


DLLFUNCTION PQbackendPID(o)                     USING STDCALL FROM libpq.dll
DLLFUNCTION PQbinaryTuples(o)                   USING STDCALL FROM libpq.dll
DLLFUNCTION PQclear(c)                          USING STDCALL FROM libpq.dll
DLLFUNCTION PQcmdStatus(o)                      USING STDCALL FROM libpq.dll
DLLFUNCTION PQcmdTuples(o)                      USING STDCALL FROM libpq.dll
DLLFUNCTION PQconnectdb(c)                      USING STDCALL FROM libpq.dll
DLLFUNCTION PQconnectPoll(u)                    USING STDCALL FROM libpq.dll
DLLFUNCTION PQconnectStart(c)                   USING STDCALL FROM libpq.dll
DLLFUNCTION PQconsumeInput(o)                   USING STDCALL FROM libpq.dll
DLLFUNCTION PQdb(o)                             USING STDCALL FROM libpq.dll
DLLFUNCTION PQerrorMessage(o)                   USING STDCALL FROM libpq.dll
DLLFUNCTION PQescapeBytea(c,n,@i)               USING STDCALL FROM libpq.dll
DLLFUNCTION PQescapeString(@b,c,l)              USING STDCALL FROM libpq.dll
DLLFUNCTION PQexec(o,c)                         USING STDCALL FROM libpq.dll
DLLFUNCTION PQexecParams(o,c,n,ct,cv,cl,cf,rf)  USING STDCALL FROM libpq.dll
DLLFUNCTION PQfinish(o)                         USING STDCALL FROM libpq.dll
DLLFUNCTION PQflush(o)                          USING STDCALL FROM libpq.dll
DLLFUNCTION PQfmod(o,i)                         USING STDCALL FROM libpq.dll
DLLFUNCTION PQfname(i, n)                       USING STDCALL FROM libpq.dll
DLLFUNCTION PQfnumber(o,c)                      USING STDCALL FROM libpq.dll
DLLFUNCTION PQfsize(o,i)                        USING STDCALL FROM libpq.dll
DLLFUNCTION PQftype(o,i)                        USING STDCALL FROM libpq.dll
DLLFUNCTION PQgetisnull(o,r,c)                  USING STDCALL FROM libpq.dll
DLLFUNCTION PQgetlength(o,r,c)                  USING STDCALL FROM libpq.dll
DLLFUNCTION PQgetResult(o)                      USING STDCALL FROM libpq.dll
DLLFUNCTION PQgetvalue(o,n,i)                   USING STDCALL FROM libpq.dll
DLLFUNCTION PQhost(o)                           USING STDCALL FROM libpq.dll
DLLFUNCTION PQisBusy(o)                         USING STDCALL FROM libpq.dll
DLLFUNCTION PQisnonblocking(o)                  USING STDCALL FROM libpq.dll
DLLFUNCTION PQmakeEmptyPGResult(o,i)            USING STDCALL FROM libpq.dll
DLLFUNCTION PQnfields(o)                        USING STDCALL FROM libpq.dll
DLLFUNCTION PQntuples(o)                        USING STDCALL FROM libpq.dll
DLLFUNCTION PQoidStatus(o)                      USING STDCALL FROM libpq.dll
DLLFUNCTION PQoidValue(o)                       USING STDCALL FROM libpq.dll
DLLFUNCTION PQoptions(o)                        USING STDCALL FROM libpq.dll
DLLFUNCTION PQpass(o)                           USING STDCALL FROM libpq.dll
DLLFUNCTION PQport(o)                           USING STDCALL FROM libpq.dll
DLLFUNCTION PQrequestcancel(o)                  USING STDCALL FROM libpq.dll
DLLFUNCTION PQreset(o)                          USING STDCALL FROM libpq.dll
DLLFUNCTION PQresetPoll(u)                      USING STDCALL FROM libpq.dll
DLLFUNCTION PQresetStart(c)                     USING STDCALL FROM libpq.dll
DLLFUNCTION PQresStatus(i)                      USING STDCALL FROM libpq.dll
DLLFUNCTION PQresultErrorMessage(o)             USING STDCALL FROM libpq.dll
DLLFUNCTION PQresultStatus(o)                   USING STDCALL FROM libpq.dll
DLLFUNCTION PQsendQuery(o,c)                    USING STDCALL FROM libpq.dll
DLLFUNCTION PQsetdbLogin(c1,c2,c3,c4,c5,c6,c7)  USING STDCALL FROM libpq.dll
DLLFUNCTION PQsetnonblocking(o,l)               USING STDCALL FROM libpq.dll
DLLFUNCTION PQsocket(o)                         USING STDCALL FROM libpq.dll
DLLFUNCTION PQstatus(o)                         USING STDCALL FROM libpq.dll
DLLFUNCTION PQtrace(o,n)                        USING STDCALL FROM libpq.dll
DLLFUNCTION PQunescapeBytea(c,n)                USING STDCALL FROM libpq.dll
DLLFUNCTION PQuntrace(o)                        USING STDCALL FROM libpq.dll
DLLFUNCTION PQuser(o)                           USING STDCALL FROM libpq.dll

